#!/usr/bin/env python

import sys
import math
import numpy as np
import matplotlib.pyplot as plt

def fitting(t,c):
  t_mean=sum(t)/len(t)
  c_log=[math.log(x) for x in c]
  c_mean=sum(c_log)/len(t)
  numer=0.0
  denom=0.0
  for i in range(len(t)):
    numer+=(t[i]*c_log[i])-(t_mean*c_mean)
    denom+=t[i]*t[i]-t_mean*t_mean
  a=numer/denom
  b=c_mean-a*t_mean
  return (a,math.exp(b))

def log10(values):
  r=[]
  missed=0
  for x in values:
    try:
      y=math.log(x)/math.log(10.0)
      for i in range(missed+1):
        r.append(y)
      missed=0
    except:
      missed+=1
  for i in range(missed):
    r.append(r[-1])
  return r

def loadFile(name,color):
  count_idx=-1
  cX_idx=-1
  t=[]
  count=[]
  cX=[]
  for l in open(name):
    l=l.split()
    if count_idx==-1:
      for s in species:
        title=s+'_'+countUnit
        if title in l:
          sys.stdout.write('%s: %s in %s\n'%(color,s,name))
          count_idx=l.index(title)
          cX_idx=l.index(s+'_mean_'+concentration)
          break
      continue
    t_value=float(l[0])/3600
    count_value=float(l[count_idx])
    cX_value=float(l[cX_idx])
    if count and count[-1]>count_value:
      sys.stdout.write('backward count at %g in %s\n'%(t_value,name))
    t.append(t_value)
    count.append(count_value)
    cX.append(cX_value)
  # (a,b)=fitting(t,count)
  # sys.stdout.write('%s --> %g * exp( %g * H )\n'%(color,b,a))
  if use_log:
    count=log10(count)
    if concentration!='pH':
      cX=log10(cX)
  return (t,count,cX,color)

files=[]
use_log=False
countUnit='fcu/g'
concentration='cL_mM/l'
species=[]
i=1
while i<len(sys.argv):
  a=sys.argv[i]
  if a=='-n':
    countUnit='fcu'
  elif a=='-log':
    use_log=True
  elif a=='-cG':
    concentration='cG_mM/l'
  elif a=='-cL':
    concentration='cL_mM/l'
  elif a=='-pH':
    concentration='pH'
  elif a=='-s' and i<len(sys.argv)-1:
    species=[sys.argv[i+1]]
    i+=1
  else:
    files.append(a)
  i+=1
if not species:
  species=['carnobacterium','lactis','listeria']

colors=['red','green','blue','grey','orange','magenta','black',
        'cyan','yellow']
expe=[]
for i in range(len(files)):
  expe.append(loadFile(files[i],colors[len(files)-1-i]))

fig,ax1=plt.subplots()
ax2=ax1.twinx()
for (t,count,cX,col) in expe:
  ax1.plot(t,cX,lw=2,linestyle='dotted',color=col)
  ax2.plot(t,count,lw=2,color=col)
ax1.set_xlabel('time (H)')
label1=concentration
label2=countUnit
if use_log:
  if concentration!='pH':
    label1='log10(%s)'%label1
  label2='log10(%s)'%label2
ax1.set_ylabel(label1)
ax2.set_ylabel(label2)
ax1.grid()
ax2.grid()
plt.tight_layout()
plt.show()

