//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "misivi.h"

#define SUSPEND_TILES 1

SimulationData sd;

ActivityResult
ActivationTile_activate(Activity *a,
                        Scheduler *s)
{
const int cpuCtx=Activity_cpuCtx(a);
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_INT_CONSTANT(cst,activation_sampling);
IMPORT_INT_CONSTANT(cst,activation_tiling);
IMPORT_INT_CONSTANT(cst,detection_sampling);
IMPORT_INT_CONSTANT(cst,detection_cell_size);
Stat * RESTRICT stat=&sd.tsd[cpuCtx].stat;
const int64_t currentStep=Scheduler_step(s);
const int fromState2=Scheduler_fromState2(s);
const int toState2=Scheduler_toState2(s);
#if SYNCHRONISE_SUBSTRATE
  const int fromState3=fromState2;
  const int toState3=toState2;
#else
  const int fromState3=Scheduler_fromState3(s);
  const int toState3=Scheduler_toState3(s);
#endif
const DetectionCellStack * RESTRICT fromStacks=sd.detectionStacks[fromState2];
DetectionCellStack *toStacks=sd.detectionStacks[toState2];
const real32_t * RESTRICT fromMasses=sd.substrate.masses[fromState2];
real32_t *toMasses=sd.substrate.masses[toState2];
Substrate * RESTRICT sub=&sd.substrate;
const real32_t * RESTRICT cG_array=sub->substances[GLUCOSE].top_c[fromState3];
real32_t *delta_cG_array=sub->substances[GLUCOSE].top_dc[toState3];
const real32_t * RESTRICT cL_array=sub->substances[LACTATE].top_c[fromState3];
real32_t *delta_cL_array=sub->substances[LACTATE].top_dc[toState3];
int lastType=-1;
BacteriumStat * RESTRICT bstat=(BacteriumStat *)0;
BacteriumActivationFnct activationFnct=(BacteriumActivationFnct)0;
const BacteriumRenderingFnct renderingFnct=
#if ENABLE_RENDERING
  sd.needsRedraw[toState2] ? &Rendering_showBacterium :
#endif
  (BacteriumRenderingFnct)0;
const bool pause=sd.pause;
ActivationTile *t=STRUCT_OF_MEMBER(a,ActivationTile,act);
const intptr_t tId=t-sd.activationTiles;
const intptr_t ty=tId/cst_activation_sampling;
const intptr_t tx=tId-ty*cst_activation_sampling;
const intptr_t cyBegin=ty*cst_activation_tiling;
const intptr_t cyEnd=(cyBegin+cst_activation_tiling>cst_detection_sampling)
                     ? cst_detection_sampling : cyBegin+cst_activation_tiling;
const intptr_t cxBegin=tx*cst_activation_tiling;
const intptr_t cxEnd=(cxBegin+cst_activation_tiling>cst_detection_sampling)
                     ? cst_detection_sampling : cxBegin+cst_activation_tiling;
// randomize to avoid bias
intptr_t cyLoopBegin,cyLoopEnd,cyLoopStep;
if(Random_int(cpuCtx,2))
  { cyLoopBegin=cyBegin; cyLoopEnd=cyEnd; cyLoopStep=1; }
else
  { cyLoopBegin=cyEnd-1; cyLoopEnd=cyBegin-1; cyLoopStep=-1; }
intptr_t cxLoopBegin,cxLoopEnd,cxLoopStep;
if(Random_int(cpuCtx,2))
  { cxLoopBegin=cxBegin; cxLoopEnd=cxEnd; cxLoopStep=1; }
else
  { cxLoopBegin=cxEnd-1; cxLoopEnd=cxBegin-1; cxLoopStep=-1; }
int foundInTile=0;
bool remainingLevels=true;
for(int level=0;remainingLevels;++level)
  {
  remainingLevels=false;
  for(intptr_t cy=cyLoopBegin;cy!=cyLoopEnd;cy+=cyLoopStep)
    {
    for(intptr_t cx=cxLoopBegin;cx!=cxLoopEnd;cx+=cxLoopStep)
      {
      const intptr_t cId=cy*cst_detection_sampling+cx;
      const DetectionCellStack * RESTRICT fromStack=fromStacks+cId;
      if((fromStack->step==currentStep)&&
         (level<fromStack->levelCount))
        {
        remainingLevels=true;
        const uint8_t *fromPage=fromStack->page;
        const DetectionCell * RESTRICT cell=
          (const DetectionCell *)(fromPage+level*cst_detection_cell_size);
        const int foundInCell=cell->count;
        if(foundInCell)
          {
          for(int i=0;i<foundInCell;++i)
            {
            const Bacterium * RESTRICT b=cell->content+i;
            int type=b->type;
            if(type!=lastType)
              {
              bstat=stat->bacteria+type;
              const BacteriumConstantRecord * RESTRICT bcst=
                cst->bacterium+type;
              activationFnct=bcst->activationFnct;
              lastType=type;
              }
            const intptr_t destTileId=
              (*activationFnct)(cell->content+i,bstat,
                                fromStacks,toStacks,
                                fromMasses,toMasses,
                                cG_array,delta_cG_array,
                                cL_array,delta_cL_array,
                                currentStep,level,cpuCtx,renderingFnct,pause);
            if(destTileId!=tId)
              {
              ActivationTile *destTile=sd.activationTiles+destTileId;
#if SUSPEND_TILES
              SpinSynchro_lockW(&destTile->sync);
              if(destTile->empty)
                {
                destTile->empty=false;
                Activity_init(&destTile->act,&ActivationTile_activate,
                              s,cpuCtx);
                }
              SpinSynchro_unlockW(&destTile->sync);
#else
              destTile->empty=false;
#endif
              }
            }
          foundInTile+=foundInCell;
          ++stat->nonEmptyCellCount;
          }
        }
      }
    }
  }
if(foundInTile)
  {
  ++stat->nonEmptyTileCount;
  for(intptr_t y=-1;y<=1;++y)
    {
    const intptr_t oty=(ty+y+cst_activation_sampling)%
                       cst_activation_sampling;
    for(intptr_t x=-1;x<=1;++x)
      {
      const intptr_t otx=(tx+x+cst_activation_sampling)%
                         cst_activation_sampling;
      const ActivationTile *otherTile=sd.activationTiles+
                                      oty*cst_activation_sampling+otx;
      if((otherTile!=t)&&(!otherTile->empty))
        {
        const int otherCpu=Activity_cpuCtx(&otherTile->act);
        Activity_adjustAffinity(&t->act,&sd.sched,otherCpu,
                                Platform_cpuProximity(cpuCtx,otherCpu));
        }
      }
    }
  }
else
  {
  // FIXME: another activity could add a Bacterium just before this test
#if SUSPEND_TILES
  SpinSynchro_lockW(&t->sync);
  t->empty=true;
  SpinSynchro_unlockW(&t->sync);
  return ACT_DEAD;
#else
  t->empty=true;
#endif
  }
return ACT_DONE;
}

void
ActivationTile_simulate(int cpuCtx,
                        void *data)
{
(void)data;
#if !SYNCHRONISE_SUBSTRATE
  if(!cpuCtx)
    {
    Substrate_beginDiffusion();
    }
#endif
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_INT_CONSTANT(cst,bacterium_type_count);
Stat * RESTRICT stat=&sd.tsd[cpuCtx].stat;
stat->nonEmptyTileCount=0;
stat->nonEmptyCellCount=0;
for(int b=0;b<cst_bacterium_type_count;++b)
  {
  BacteriumStat * RESTRICT bstat=stat->bacteria+b;
  bstat->count=0;
  bstat->cG=0.0;
  bstat->cL=0.0;
  bstat->pH=0.0;
  bstat->gamma=0.0;
  bstat->starving=0.0;
  bstat->detection=0.0;
  }
Scheduler_activate(cpuCtx,&sd.sched);
}

void
prepareActivation(int cpuCtx,
                  void *data)
{
(void)data;
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_INT_CONSTANT(cst,substrate_sampling);
const int toState2=Scheduler_toState2(&sd.sched);
#if SYNCHRONISE_SUBSTRATE
  const int toState3=toState2;
#else
  const int toState3=Scheduler_toState3(&sd.sched);
#endif
Substrate * RESTRICT sub=&sd.substrate;
real32_t * RESTRICT masses=sub->masses[toState2];
const intptr_t topCellCount=cst_substrate_sampling*cst_substrate_sampling;
const int cpuCount=Platform_cpuCtxCount();
const intptr_t topCellBegin=Platform_splitBegin(0,topCellCount,
                                                cpuCount,cpuCtx);
const intptr_t topCellEnd=Platform_splitEnd(0,topCellCount,
                                            cpuCount,cpuCtx);
TRY_TO_VECTORISE
for(intptr_t id=topCellBegin;id<topCellEnd;++id)
  {
  masses[id]=0.0f;
  }
for(int i=0;i<SUBSTANCE_COUNT;++i)
  {
  Substance * RESTRICT s=sub->substances+i;
  real32_t * RESTRICT top_dc=s->top_dc[toState3];
  TRY_TO_VECTORISE
  for(intptr_t id=topCellBegin;id<topCellEnd;++id)
    {
    top_dc[id]=0.0f;
    }
  }
}

void
createTiles(int cpuCtx,
            void *data)
{
(void)data;
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_INT_CONSTANT(cst,activation_sampling);
ActivationTile * RESTRICT tiles=sd.activationTiles;
const intptr_t tileCount=cst_activation_sampling*cst_activation_sampling;
const int cpuCount=Platform_cpuCtxCount();
const intptr_t tileBegin=Platform_splitBegin(0,tileCount,cpuCount,cpuCtx);
const intptr_t tileEnd=Platform_splitEnd(0,tileCount,cpuCount,cpuCtx);
for(intptr_t tId=tileBegin;tId<tileEnd;++tId)
  {
  ActivationTile * RESTRICT t=tiles+tId;
  SpinSynchro_init(&t->sync);
  t->empty=false;
  Activity_init(&t->act,&ActivationTile_activate,&sd.sched,cpuCtx);
  }
}

void
createBacteria(int cpuCtx,
               void *data)
{
(void)data;
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_INT_CONSTANT(cst,bacterium_type_count);
IMPORT_R32_CONSTANT(cst,substrate_half_size);
const int64_t nextStep=Scheduler_step(&sd.sched)+1;
const int toState2=Scheduler_toState2(&sd.sched);
DetectionCellStack * RESTRICT toStacks=sd.detectionStacks[toState2];
real32_t * RESTRICT toMasses=sd.substrate.masses[toState2];
const int cpuCount=Platform_cpuCtxCount();
for(int b=0;b<cst_bacterium_type_count;++b)
  {
  const BacteriumConstantRecord * RESTRICT bcst=cst->bacterium+b;
  IMPORT_INT_CONSTANT(bcst,initial_count);
  IMPORT_R32_CONSTANT(bcst,distribution_center_x);
  IMPORT_R32_CONSTANT(bcst,distribution_center_y);
  IMPORT_R32_CONSTANT(bcst,distribution_inner_crown);
  IMPORT_R32_CONSTANT(bcst,distribution_outer_crown);
  IMPORT_R32_CONSTANT(bcst,mitosis_consumption);
  const BacteriumCreationFnct creationFnct=bcst->creationFnct;
  const real32_t dMin=fminf(bcst_distribution_inner_crown,
                            bcst_distribution_outer_crown);
  const real32_t dMax=fmaxf(bcst_distribution_inner_crown,
                            bcst_distribution_outer_crown);
  const real32_t dMin2=dMin*dMin;
  const real32_t dRange2=dMax*dMax-dMin2;
  const intptr_t begin=Platform_splitBegin(0,bcst_initial_count,
                                           cpuCount,cpuCtx);
  const intptr_t end=Platform_splitEnd(0,bcst_initial_count,
                                       cpuCount,cpuCtx);
  for(intptr_t i=begin;i<end;++i)
    {
    real32_t x,y;
    do
      {
      const real32_t rnd=dMin2+dRange2*Random_real32(cpuCtx);
      polarToCart2D32(sqrtf(rnd),Random_real32(cpuCtx)*M_2PIf,&x,&y);
      x=(x+bcst_distribution_center_x)*cst_substrate_half_size;
      y=(y+bcst_distribution_center_y)*cst_substrate_half_size;
      } while((x<=-cst_substrate_half_size)||
              (x>=cst_substrate_half_size)||
              (y<=-cst_substrate_half_size)||
              (y>=cst_substrate_half_size));
    // initialise uniformly along mitosis so that this start condition is
    // considered when fitting with an exponential function
    const real32_t distribution=(1.0f+2.0f*i)/(2.0f*bcst_initial_count);
    const real32_t glucose=bcst_mitosis_consumption*distribution;
    const int level=0;
    (*creationFnct)(b,x,y,glucose,toStacks,toMasses,
                    nextStep,level,cpuCtx);
    }
  }
}

void
SimulationData_init(void)
{
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_INT_CONSTANT(cst,detection_sampling);
IMPORT_INT_CONSTANT(cst,activation_sampling);
IMPORT_INT_CONSTANT(cst,bacterium_type_count);
Scheduler_init(&sd.sched);
Scheduler_setWorkStealingPeriod(&sd.sched,8);
const int64_t nextStep=Scheduler_step(&sd.sched)+1;
sd.mustQuit=false;
sd.pause=false;
sd.needsSubstrateReset=true;
#if ENABLE_RENDERING
  sd.needsRedraw[0]=sd.needsRedraw[1]=false;
#endif
const intptr_t cellCount=cst_detection_sampling*cst_detection_sampling;
for(int s=0;s<2;++s)
  {
  intptr_t size=cellCount*sizeof(DetectionCellStack);
  DetectionCellStack * RESTRICT stacks=(DetectionCellStack *)
                                       Memory_pageAlloc(&size,true);
  sd.detectionStacks[s]=stacks;
  for(intptr_t i=0;i<cellCount;++i)
    {
    DetectionCellStack * RESTRICT stack=stacks+i;
    SpinSynchro_init(&stack->sync);
    stack->levelCount=0;
    stack->step=nextStep;
    stack->pageSize=0;
    stack->page=(uint8_t *)0;
    }
  }
const intptr_t tileCount=cst_activation_sampling*cst_activation_sampling;
intptr_t size=tileCount*sizeof(ActivationTile);
sd.activationTiles=(ActivationTile *)Memory_pageAlloc(&size,true);
Platform_parallel(&createTiles,(void *)0);
for(int i=0,i_end=Platform_cpuCtxCount();i<i_end;++i)
  {
  ThreadSimulationData * RESTRICT tsd=sd.tsd+i;
  Stat * RESTRICT stat=&tsd->stat;
  stat->nonEmptyTileCount=0;
  stat->nonEmptyCellCount=0;
  for(int b=0;b<cst_bacterium_type_count;++b)
    {
    BacteriumStat * RESTRICT bstat=stat->bacteria+b;
    bstat->count=0;
    bstat->cG=0.0;
    bstat->cL=0.0;
    bstat->pH=0.0;
    bstat->gamma=0.0;
    bstat->starving=0.0;
    bstat->detection=0.0;
    }
  }
}

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
const char *constantFileName="misivi.cst";
bool displayDebug=false;
for(int i=1,i_end=Platform_argc();i<i_end;++i)
  {
  if(!strcmp(Platform_argv(i),"-d"))
    {
    displayDebug=true;
    }
  else
    {
    constantFileName=Platform_argv(i);
    }
  }
Platform_msg("loading simulation parameters from \"%s\"\n",
             constantFileName);
ConstantRecord_init(constantFileName);
SimulationData_init();
#if ENABLE_RENDERING
  Rendering_init();
#endif
Substrate_init();
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_R32_CONSTANT(cst,substrate_step);
IMPORT_R32_CONSTANT(cst,substrate_size);
IMPORT_R32_CONSTANT(cst,substrate_thickness);
IMPORT_R32_CONSTANT(cst,substrate_density);
IMPORT_INT_CONSTANT(cst,substrate_plane_count);
IMPORT_INT_CONSTANT(cst,substrate_sampling);
IMPORT_R32_CONSTANT(cst,substrate_plane_progression);
IMPORT_R32_CONSTANT(cst,dt);
IMPORT_R32_CONSTANT(cst,detection_step);
IMPORT_INT_CONSTANT(cst,detection_sampling);
IMPORT_R32_CONSTANT(cst,activation_step);
IMPORT_INT_CONSTANT(cst,activation_sampling);
IMPORT_INT_CONSTANT(cst,activation_tiling);
IMPORT_INT_CONSTANT(cst,maximal_bacterium_count);
IMPORT_R32_CONSTANT(cst,maximal_simulation_duration);
IMPORT_STR_CONSTANT(cst,log_file_name);
IMPORT_R32_CONSTANT(cst,log_period);
IMPORT_INT_CONSTANT(cst,substrate_dt_divisor);
IMPORT_INT_CONSTANT(cst,bacterium_type_count);
const real32_t perLitre=1e-3f/(cst_substrate_thickness*
                               cst_substrate_size*cst_substrate_size);
const real32_t perGram=perLitre/cst_substrate_density;
Platform_msg("substrate step: %g [m]\n",
             cst_substrate_step);
Platform_msg("substrate size: %g x %g x %g [m], %g g\n",
             cst_substrate_size,
             cst_substrate_size,
             cst_substrate_thickness,
             1.0f/perGram);
Platform_msg("substrate grid: %d x %d x %d (%g depth progression)\n",
             cst_substrate_sampling,
             cst_substrate_sampling,
             cst_substrate_plane_count,
             cst_substrate_plane_progression);
Platform_msg("detection step: %g [m]\n",
             cst_detection_step);
Platform_msg("detection grid: %d x %d\n",
             cst_detection_sampling,
             cst_detection_sampling);
Platform_msg("activation step: %g [m] (%d x detection step)\n",
             cst_activation_step,
             cst_activation_tiling);
Platform_msg("activation grid: %d x %d\n",
             cst_activation_sampling,
             cst_activation_sampling);
Platform_msg("dt: %g [s] (%g for substrate)\n",
             cst_dt,
             cst_dt/cst_substrate_dt_divisor);
for(int b=0;b<cst_bacterium_type_count;++b)
  {
  const BacteriumConstantRecord * RESTRICT bcst=cst->bacterium+b;
  IMPORT_STR_CONSTANT(bcst,name);
  IMPORT_R32_CONSTANT(bcst,radius_max);
  real32_t minRadius=Bacterium_radiusFromGlucose(b,0.0f);
  Platform_msg("%s radius: %g to %g [m]\n",
               bcst_name,
               minRadius,
               bcst_radius_max);
  }
Platform_parallel(&createBacteria,(void *)0);
FILE *logFile=NULL;
if(cst_log_file_name[0])
  {
  Platform_msg("writing simulation results to \"%s\" every %g seconds\n",
               cst_log_file_name,
               cst_log_period);
  logFile=Path_fopen(cst_log_file_name,"w");
  if(!logFile)
    {
    char msg[0x100];
    snprintf(msg,0x100,"%s open failure",cst_log_file_name);
    SYS_DROP_DEAD(msg);
    }
  fprintf(logFile,"time_s\ttime_dhms\telapsed_s\telapsed_dhms");
  for(int b=0;b<cst_bacterium_type_count;++b)
    {
    const BacteriumConstantRecord * RESTRICT bcst=cst->bacterium+b;
    IMPORT_INT_CONSTANT(bcst,initial_count);
    if(bcst_initial_count)
      {
      IMPORT_STR_CONSTANT(bcst,name);
      fprintf(logFile,"\t%s_fcu"
                      "\t%s_fcu/g"
                      "\t%s_mean_cG_mM/l"
                      "\t%s_mean_cL_mM/l"
                      "\t%s_mean_pH"
                      "\t%s_mean_gamma"
                      "\t%s_starving_ratio",
                      bcst_name,
                      bcst_name,
                      bcst_name,
                      bcst_name,
                      bcst_name,
                      bcst_name,
                      bcst_name);
      }
    }
  fprintf(logFile,"\n");
  fflush(logFile);
  }
int64_t lastLog=-1;
real64_t elapsed=0.0;
int64_t now=Platform_time();
int perfCount=0;
int64_t lastPerf=now;
#if ENABLE_RENDERING
  int64_t lastDraw=now;
#endif
int64_t simulationSteps=0;
while(!sd.mustQuit)
  {
  int64_t deltaElapsed=Platform_time()-now;
  now+=deltaElapsed;
  Scheduler_advanceStep(&sd.sched);
  if(sd.needsSubstrateReset)
    {
    Substrate_reset();
    sd.needsSubstrateReset=false;
    }
  Platform_parallel(&prepareActivation,(void *)0);
#if ENABLE_RENDERING
  const int64_t drawDt=now-lastDraw;
  const int toState2=Scheduler_toState2(&sd.sched);
  sd.needsRedraw[toState2]=drawDt>=100000;
  if(sd.needsRedraw[toState2]) lastDraw+=drawDt;
  Platform_parallel(&Rendering_activate,(void *)0);
#else
  Platform_parallel(&ActivationTile_simulate,(void *)0);
#endif
#if SYNCHRONISE_SUBSTRATE
  Substrate_diffuse();
#else
  Substrate_endDiffusion();
#endif
  if(sd.pause)
    { continue; }
  elapsed+=deltaElapsed*1e-6;
  int elapsedDay=0;
  int elapsedHour=0;
  int elapsedMin=0;
  int elapsedSec=0;
  const real64_t simulationTime=(++simulationSteps)*cst_dt;
  int dayCount=0;
  int hourCount=0;
  int minCount=0;
  real64_t secCount=0.0;
  intptr_t totalBacteriumCount=0;
  Stat reducedStat;
  real64_t perf=0.0;
  bool logStats=false;
  bool displayStats=false;
  if(logFile)
    {
    const int64_t newLog=(int64_t)floor(simulationTime/cst_log_period);
    if(lastLog!=newLog)
      {
      lastLog=newLog;
      logStats=true;
      }
    }
  ++perfCount;
  const int64_t perfDt=now-lastPerf;
  if(perfDt>=2000000)
    {
    lastPerf+=perfDt;
    perf=perfCount/(1e-6*perfDt);
    perfCount=0;
    displayStats=true;
    }
  if(logStats||displayStats) // compute stats
    {
    elapsedDay=(int)(elapsed/86400.0);
    elapsedHour=(int)fmod(elapsed/3600.0,24.0);
    elapsedMin=(int)fmod(elapsed/60.0,60.0);
    elapsedSec=(int)fmod(elapsed,60.0);
    dayCount=(int)(simulationTime/86400.0);
    hourCount=(int)fmod(simulationTime/3600.0,24.0);
    minCount=(int)fmod(simulationTime/60.0,60.0);
    secCount=fmod(simulationTime,60.0);
    totalBacteriumCount=0;
    for(int b=0;b<cst_bacterium_type_count;++b)
      {
      BacteriumStat * RESTRICT bstat=reducedStat.bacteria+b;
      bstat->count=0;
      bstat->cG=0.0;
      bstat->cL=0.0;
      bstat->pH=0.0;
      bstat->gamma=0.0;
      bstat->starving=0.0;
      bstat->detection=0.0;
      for(int i=0,i_end=Platform_cpuCtxCount();i<i_end;++i)
        {
        const ThreadSimulationData * RESTRICT tsd=sd.tsd+i;
        const BacteriumStat * RESTRICT tsdStat=tsd->stat.bacteria+b;
        bstat->count+=tsdStat->count;
        bstat->cG+=tsdStat->cG;
        bstat->cL+=tsdStat->cL;
        bstat->pH+=tsdStat->pH;
        bstat->gamma+=tsdStat->gamma;
        bstat->starving+=tsdStat->starving;
        bstat->detection+=tsdStat->detection;
        }
      const real64_t div=1.0/ONE_IF_ZERO(bstat->count);
      bstat->cG*=div;
      bstat->cL*=div;
      bstat->pH*=div;
      bstat->gamma*=div;
      bstat->starving*=div;
      bstat->detection*=div;
      totalBacteriumCount+=bstat->count;
      }
    reducedStat.nonEmptyTileCount=0;
    reducedStat.nonEmptyCellCount=0;
    for(int i=0,i_end=Platform_cpuCtxCount();i<i_end;++i)
      {
      Stat * RESTRICT stat=&sd.tsd[i].stat;
      reducedStat.nonEmptyTileCount+=stat->nonEmptyTileCount;
      reducedStat.nonEmptyCellCount+=stat->nonEmptyCellCount;
      }
    }
  if(logStats)
    {
    fprintf(logFile,"%g\t%dd:%dh:%dm:%gs\t%g\t%dd:%dh:%dm:%ds",
                    simulationTime,
                    dayCount,hourCount,minCount,secCount,
                    elapsed,
                    elapsedDay,elapsedHour,elapsedMin,elapsedSec);
    for(int b=0;b<cst_bacterium_type_count;++b)
      {
      const BacteriumConstantRecord * RESTRICT bcst=cst->bacterium+b;
      IMPORT_INT_CONSTANT(bcst,initial_count);
      if(bcst_initial_count)
        {
        const BacteriumStat * RESTRICT bstat=reducedStat.bacteria+b;
        fprintf(logFile,"\t%"FPTR"d"
                        "\t%g"
                        "\t%g"
                        "\t%g"
                        "\t%g"
                        "\t%g"
                        "\t%g",
                        bstat->count,
                        bstat->count*perGram,
                        bstat->cG*1e3f,
                        bstat->cL*1e3f,
                        bstat->pH,
                        bstat->gamma,
                        bstat->starving);
        }
      }
    fprintf(logFile,"\n");
    fflush(logFile);
    }
  if(displayStats)
    {
    Platform_msg("%dd:%dh:%dm:%gs (%g Hz, %dd:%dh:%dm:%ds elapsed)\n",
                 dayCount,hourCount,minCount,secCount,
                 perf,
                 elapsedDay,elapsedHour,elapsedMin,elapsedSec);
    for(int b=0;b<cst_bacterium_type_count;++b)
      {
      const BacteriumConstantRecord * RESTRICT bcst=cst->bacterium+b;
      IMPORT_INT_CONSTANT(bcst,initial_count);
      if(bcst_initial_count)
        {
        IMPORT_STR_CONSTANT(bcst,name);
        const BacteriumStat * RESTRICT bstat=reducedStat.bacteria+b;
        Platform_msg("  %s: %"FPTR"d (%g/g)"
                     " %.2f%% gamma, %.2f%% starving\n"
                     "    %g cG [mM/l], %g cL [mM/l], %.2f pH\n",
                     bcst_name,
                     bstat->count,
                     bstat->count*perGram,
                     100.0*bstat->gamma,
                     100.0*bstat->starving,
                     bstat->cG*1e3,
                     bstat->cL*1e3,
                     bstat->pH);
        }
      }
    if(displayDebug)
      {
      Platform_msg("  %"FPTR"d tiles (%g bacteria),"
                   " %"FPTR"d cells (%g bacteria)\n",
                   reducedStat.nonEmptyTileCount,
                   totalBacteriumCount/
                   (real64_t)ONE_IF_ZERO(reducedStat.nonEmptyTileCount),
                   reducedStat.nonEmptyCellCount,
                   totalBacteriumCount/
                   (real64_t)ONE_IF_ZERO(reducedStat.nonEmptyCellCount));
      Platform_msg("  detection:");
      for(int b=0;b<cst_bacterium_type_count;++b)
        {
        const BacteriumConstantRecord * RESTRICT bcst=cst->bacterium+b;
        IMPORT_INT_CONSTANT(bcst,initial_count);
        if(bcst_initial_count)
          {
          const BacteriumStat * RESTRICT bstat=reducedStat.bacteria+b;
          Platform_msg(" %g",
                       bstat->detection);
          }
        }
      Platform_msg("\n");
      }
    }
  if(cst_maximal_bacterium_count&&
     (totalBacteriumCount>cst_maximal_bacterium_count))
    {
    Platform_msg("AUTOMATIC TERMINATION: %"FPTR"d bacteria reached\n",
                 totalBacteriumCount);
    sd.mustQuit=true;
    }
  if(cst_maximal_simulation_duration&&
     (simulationTime>cst_maximal_simulation_duration))
    {
    Platform_msg("AUTOMATIC TERMINATION: %gs of simulation reached\n",
                 simulationTime);
    sd.mustQuit=true;
    }
  }
CudaStream_hostSync(&sd.substrate.compStream);
CudaStream_hostSync(&sd.substrate.sendStream);
CudaStream_hostSync(&sd.substrate.recvStream);
Cuda_profiling(false);
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
